"""Containers for Work-in-progress data."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Callable

import click

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from collections.abc import Generator
    from pathlib import Path

    from .data import PoodleConfig
    from .interfaces import Mutator


class PoodleWork:
    """Work-in-progress data gathered into single structure for easy passing between functions."""

    def __init__(self, config: PoodleConfig) -> None:
        """Init from PoodleConfig."""
        self.config = config
        self.folder_zips: dict[Path, Path] = {}
        self.mutators: list[Mutator | Callable] = []
        self.runner: Callable = lambda *_, **__: None
        self.reporters: list[Callable] = []

        self.echo = click.echo if config.echo_enabled else lambda *_, **__: None

        def number_generator() -> Generator[int, Any, None]:
            i = 1
            while True:
                yield i
                i += 1

        self._num_gen = number_generator()

    def next_num(self) -> str:
        """Return the next value from Sequence as a string."""
        return str(next(self._num_gen))



"""Various utility functions."""

from __future__ import annotations

import logging
import re
from io import StringIO
from pprint import pprint
from typing import TYPE_CHECKING, Any
from zipfile import ZipFile

if TYPE_CHECKING:
    from pathlib import Path

    from .data_types import PoodleWork

logger = logging.getLogger(__name__)


def files_list_for_folder(glob: str, filter_regex: list[str], folder: Path) -> list[Path]:
    """Retrieve list of all files in specified folder.

    Search recursively for files matching glob.
    Remove files matching any of the filter_regex values.
    """
    logger.debug("files_list_for_folder glob=%s filter_regex=%s folder=%s", glob, filter_regex, folder)

    files = list(folder.rglob(glob))

    for regex in filter_regex:
        files = [file for file in files if not any(re.search(regex, part) for part in file.parts)]

    logger.debug("files_list_for_folder results: folder=%s files=%s", folder, files)
    return files


def files_list_for_source_folders(work: PoodleWork) -> dict[Path, list[Path]]:
    """Build map of Folder to all files in folder to include in zips."""
    return {
        folder: files_list_for_folder(
            "*",
            work.config.file_copy_filters,
            folder,
        )
        for folder in work.config.source_folders
    }


def create_temp_zips(work: PoodleWork) -> None:
    """Create a temporary zip file for each folder in source_folders."""
    work.config.work_folder.mkdir(parents=True, exist_ok=True)
    for folder, files in files_list_for_source_folders(work).items():
        zip_file = work.config.work_folder / ("src-" + work.next_num() + ".zip")
        logger.info("Creating zip file: %s", zip_file)
        work.folder_zips[folder] = zip_file
        with ZipFile(zip_file, "w") as target_zip:
            for file in files:
                target_zip.write(file)


def dynamic_import(object_to_import: str) -> Any:  # noqa: ANN401
    """Import an Object represented by provided python name."""
    logger.debug("Import object: %s", object_to_import)
    parts = object_to_import.split(".")
    module_str = ".".join(parts[:-1])
    obj_def = parts[-1:][0]
    module = __import__(module_str, fromlist=[obj_def])
    return getattr(module, obj_def)


def pprint_str(obj: Any) -> str:  # noqa: ANN401
    """Pretty Print an object to a string."""
    out = StringIO()
    pprint(obj, stream=out, width=150)  # noqa: T203
    return out.getvalue()



"""Command Line Interface."""

from __future__ import annotations

import sys
from pathlib import Path

import click

from . import PoodleInputError, core
from .config import build_config

CONTEXT_SETTINGS = {
    "max_content_width": 200,
}


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("source", type=click.Path(exists=True, path_type=Path), nargs=-1)
@click.option("-C", "--config_file", help="Configuration File.", type=click.Path(exists=True, path_type=Path))
@click.option("-q", "verbosity", help="Quiet mode: disabled normal output, and loglevel=ERROR", flag_value="q")
@click.option("-v", "verbosity", help="Verbose mode: loglevel=INFO", flag_value="v")
@click.option("-vv", "verbosity", help="Very Verbose mode: loglevel=DEBUG", flag_value="vv")
@click.option("-P", "--max_workers", type=int, help="Maximum number of parallel workers.")
@click.option(
    "--exclude", "excludes", multiple=True, help="Add a regex filter for which files NOT to mutate.  Multiple allowed."
)
@click.option(
    "--only",
    "only_files",
    multiple=True,
    help="Glob pattern for files to mutate.  If specified, no other files will be mutated.  Multiple allowed.",
)
# @click.option("-R", "--runner", help="Runner Name or Module Name for runner to use")
def main(
    source: tuple[Path],
    config_file: Path | None,
    verbosity: str | None,
    max_workers: int | None,
    excludes: tuple[str],
    only_files: tuple[str],
) -> None:
    """Poodle Mutation Test Tool."""
    try:
        config = build_config(source, config_file, verbosity, max_workers, excludes, only_files)
    except PoodleInputError as err:
        click.echo(err.args)
        sys.exit(4)

    try:
        core.main(config)
    except KeyboardInterrupt:
        click.echo("Aborted due to Keyboard Interrupt!")
        sys.exit(2)


# pytest return codes
# Exit code 0: All tests were collected and passed successfully
# Exit code 1: Tests were collected and run but some of the tests failed
# Exit code 2: Test execution was interrupted by the user
# Exit code 3: Internal error happened while executing tests
# Exit code 4: pytest command line usage error
# Exit code 5: No tests were collected

if __name__ == "__main__":
    main()



from pathlib import Path
from unittest import mock

import pytest

from poodle import util
from poodle.data_types import PoodleWork
from tests.data_types.test_data import PoodleConfigStub


@pytest.fixture()
def mock_logger():
    with mock.patch("poodle.util.logger") as logger:
        yield logger


def test_logger():
    assert util.logger.name == "poodle.util"


class TestFilesList:
    def test_files_list_for_folder(self, mock_logger):
        mock_folder = mock.MagicMock()
        mock_folder.rglob.return_value = (
            Path("src/project/file_1.py"),
            Path("src/project/file_test_2.py"),
            Path("src/project/__pycache__/file_1.pyc"),
            Path("src/project/__pycache__/file_test_2.pyc"),
        )
        files = util.files_list_for_folder("*.py", [r"^test_.*\.py$", r"_test\.py$", r"^__pycache__$"], mock_folder)
        assert files == [
            Path("src/project/file_1.py"),
            Path("src/project/file_test_2.py"),
        ]
        mock_folder.rglob.assert_called_with("*.py")
        mock_logger.debug.assert_any_call(
            "files_list_for_folder glob=%s filter_regex=%s folder=%s",
            "*.py",
            [r"^test_.*\.py$", r"_test\.py$", r"^__pycache__$"],
            mock_folder,
        )
        mock_logger.debug.assert_any_call(
            "files_list_for_folder results: folder=%s files=%s",
            mock_folder,
            files,
        )

    def test_files_list_for_source_folders(self, mock_logger):
        mock_folder_1 = mock.MagicMock()
        mock_folder_1.rglob.return_value = (
            Path("file_1.py"),
            Path("test_file_1.py"),
        )
        mock_folder_2 = mock.MagicMock()
        mock_folder_2.rglob.return_value = (
            Path("file_test_2.py"),
            Path("file_2_test.py"),
        )
        work = PoodleWork(
            config=PoodleConfigStub(
                source_folders=[mock_folder_1, mock_folder_2],
                file_copy_filters=[r"^test_.*\.py$", r"_test\.py$"],
            )
        )
        assert util.files_list_for_source_folders(work) == {
            mock_folder_1: [Path("file_1.py")],
            mock_folder_2: [Path("file_test_2.py")],
        }
        mock_folder_1.rglob.assert_called_with("*")
        mock_folder_2.rglob.assert_called_with("*")
        mock_logger.debug.assert_has_calls(
            [
                mock.call(
                    "files_list_for_folder glob=%s filter_regex=%s folder=%s",
                    "*",
                    [r"^test_.*\.py$", r"_test\.py$"],
                    mock_folder_1,
                ),
                mock.call(
                    "files_list_for_folder results: folder=%s files=%s",
                    mock_folder_1,
                    [Path("file_1.py")],
                ),
                mock.call(
                    "files_list_for_folder glob=%s filter_regex=%s folder=%s",
                    "*",
                    [r"^test_.*\.py$", r"_test\.py$"],
                    mock_folder_2,
                ),
                mock.call(
                    "files_list_for_folder results: folder=%s files=%s",
                    mock_folder_2,
                    [Path("file_test_2.py")],
                ),
            ]
        )

    @mock.patch("poodle.util.files_list_for_source_folders")
    @mock.patch("poodle.util.ZipFile")
    def test_create_temp_zips(self, ZipFile, files_list_for_source_folders, mock_logger):  # noqa: N803
        work_folder = mock.MagicMock()
        work_folder.__truediv__.side_effect = lambda x: Path(x)
        files_list_for_source_folders.return_value = {
            Path("example_1"): [Path("file_1.py")],
            Path("example_2"): [Path("file_2.py")],
        }
        work = PoodleWork(
            config=PoodleConfigStub(
                work_folder=work_folder,
            )
        )
        util.create_temp_zips(work)
        work_folder.mkdir.assert_called_with(parents=True, exist_ok=True)
        files_list_for_source_folders.assert_called_with(work)
        mock_logger.info.assert_has_calls(
            [
                mock.call("Creating zip file: %s", Path("src-1.zip")),
                mock.call("Creating zip file: %s", Path("src-2.zip")),
            ]
        )
        assert work.folder_zips == {
            Path("example_1"): Path("src-1.zip"),
            Path("example_2"): Path("src-2.zip"),
        }
        ZipFile.assert_any_call(Path("src-1.zip"), "w")
        ZipFile.assert_any_call(Path("src-2.zip"), "w")
        target_zip = ZipFile.return_value.__enter__.return_value
        target_zip.write.assert_has_calls(
            [
                mock.call(Path("file_1.py")),
                mock.call(Path("file_2.py")),
            ]
        )


class TestDynamicImport:
    def test_dynamic_import(self, mock_logger):
        runner = util.dynamic_import("poodle.runners.command_line.runner")
        import poodle.runners.command_line

        assert runner is poodle.runners.command_line.runner

        mock_logger.debug.assert_called_with("Import object: %s", "poodle.runners.command_line.runner")


class TestPrint:
    def test_pprint_str(self):
        assert util.pprint_str({"key1": "value1", "key2": "value2"}) == "{'key1': 'value1', 'key2': 'value2'}\n"

    @mock.patch("poodle.util.StringIO")
    @mock.patch("poodle.util.pprint")
    def test_pprint_str_mock(self, pprint, StringIO):  # noqa: N803
        obj = {"key1": "value1", "key2": "value2"}
        util.pprint_str(obj)
        out = StringIO.return_value
        pprint.assert_called_with(obj, stream=out, width=150)
